"""Run MCMC for retrieval"""

"""___Built-In Modules___"""

"""___Third-Party Modules___"""
import os

import numpy as np
from scipy.optimize import minimize

import punpy.utilities.utilities as util

os.environ["OMP_NUM_THREADS"] = "1"

"""___NPL Modules___"""

"""___Authorship___"""
__author__ = "Pieter De Vis"
__created__ = "01/03/2020"
__maintainer__ = "Pieter De Vis"
__email__ = "pieter.de.vis@npl.co.uk"
__status__ = "Development"

inds_cache = {}


# lock = threading.Lock()
class LPURetrieval:
    def __init__(
        self,
        measurement_function,
        observed,
        syst_uncertainty=None,
        rand_uncertainty=None,
        cov=None,
        uplims=+np.inf,
        downlims=-np.inf,
        Jx=None,
    ):
        """
        Initialise Law of Propagation of Uncertainty retrieval

        :param measurement_function:
        :type measurement_function:
        :param observed:
        :type observed:
        :param syst_uncertainty:
        :type syst_uncertainty:
        :param rand_uncertainty:
        :type rand_uncertainty:
        :param cov:
        :type cov:
        :param uplims:
        :type uplims:
        :param downlims:
        :type downlims:
        """
        self.measurement_function = measurement_function
        self.observed = observed
        self.rand_uncertainty = np.array([rand_uncertainty])
        self.syst_uncertainty = np.array([syst_uncertainty])
        if cov is None:
            self.invcov = cov
        else:
            self.invcov = np.linalg.inv(np.ascontiguousarray(cov))
            # print(observed,cov,self.invcov)
        self.Jx = Jx
        self.uplims = np.array(uplims)
        self.downlims = np.array(downlims)

    def run_retrieval(self, theta_0, return_corr=True):
        res = minimize(self.find_chisum, theta_0)
        if self.Jx is None:
            Jx = util.calculate_Jacobian(self.measurement_function, res.x)
        else:
            Jx = self.Jx
        # print("wer",res.x,theta_0,Jx,util.calculate_Jacobian(self.measurement_function,theta_0))

        return tuple(res.x) + tuple(self.process_inverse_jacobian(Jx, return_corr))

    def process_inverse_jacobian(self, J, return_corr=True):
        # print(
        #     self.invcov,
        #     J.T,
        #     np.dot(J.T, self.invcov),
        #     np.dot(np.dot(J.T, self.invcov), J),
        #     np.linalg.inv(np.dot(np.dot(J.T, self.invcov), J)),
        # )
        covx = np.linalg.inv(np.dot(np.dot(J.T, self.invcov), J))
        u_func = np.diag(covx) ** 0.5
        corr_x = util.convert_cov_to_corr(covx, u_func)
        if return_corr:
            return u_func, corr_x
        else:
            return u_func
        # if not return_corr:
        #     return u_func.reshape(shape_y)
        # else:
        #     if output_vars == 1:
        #         return u_func.reshape(shape_y),corr_y
        #     else:
        #         #create an empty arrays and then populate it with the correlation matrix for each output parameter individually
        #         corr_ys = np.empty(output_vars,dtype=object)
        #         for i in range(output_vars):
        #             corr_ys[i] = corr_y[int(i*len(corr_y)/output_vars):
        #                                 int((i+1)*len(corr_y)/output_vars),
        #                          int(i*len(corr_y)/output_vars):
        #                          int((i+1)*len(corr_y)/output_vars)]
        #
        #         # #calculate correlation matrix between the different outputs produced by the measurement function.
        #         # corr_out=np.corrcoef(MC_y.reshape((output_vars,-1)))
        #
        #         return u_func.reshape(shape_y),corr_ys  #,corr_out

    def find_chisum(self, theta):
        model = self.measurement_function(theta)
        diff = model - self.observed
        if np.isfinite(np.sum(diff)):
            if self.invcov is None:
                return np.sum((diff) ** 2 / self.rand_uncertainty ** 2)
            else:
                # print(diff,np.linalg.inv(self.cov),np.dot(np.dot(diff.T,self.invcov),diff))
                return np.dot(np.dot(diff.T, self.invcov), diff)
        else:
            return np.inf



"""Use Monte Carlo to propagate uncertainties"""

import warnings
from multiprocessing import Pool

import numpy as np

import punpy.utilities.utilities as util

"""___Authorship___"""
__author__ = "Pieter De Vis"
__created__ = "30/03/2019"
__maintainer__ = "Pieter De Vis"
__email__ = "pieter.de.vis@npl.co.uk"
__status__ = "Development"


class MCPropagation:
    def __init__(self, steps, parallel_cores=0):
        """
        Initialise MC Propagator

        :param steps: number of MC iterations
        :type steps: int
        :param parallel_cores: number of CPU to be used in parallel processing
        :type parallel_cores: int
        """

        self.MCsteps = steps
        self.parallel_cores = parallel_cores

    def propagate_random(
        self,
        func,
        x,
        u_x,
        corr_x=None,
        param_fixed=None,
        corr_between=None,
        return_corr=False,
        return_samples=False,
        repeat_dims=-99,
        corr_axis=-99,
        fixed_corr_var=False,
        output_vars=1,
        PD_corr=True,
    ):
        """
        Propagate random uncertainties through measurement function with n input quantities.
        Input quantities can be floats, vectors (1d-array) or images (2d-array).
        Random uncertainties arise when there is no correlation between repeated measurements.
        It is possible (though rare) that there is a correlation in one of the dimensions that is not one of the repeat_dims.

        :param func: measurement function
        :type func: function
        :param x: list of input quantities (usually numpy arrays)
        :type x: list[array]
        :param u_x: list of random uncertainties on input quantities (usually numpy arrays)
        :type u_x: list[array]
        :param corr_x: list of correlation matrices (n,n) along non-repeating axis, defaults to None. Can optionally be set to "rand" (diagonal correlation matrix), "syst" (correlation matrix of ones) or a custom correlation matrix.
        :type corr_x: list[array], optional
        :param param_fixed: when repeat_dims>=0, set to true or false to indicate for each input quantity whether it has repeated measurements that should be split (param_fixed=False) or whether the input is fixed (param fixed=True), defaults to None (no inputs fixed).
        :type param_fixed: list of bools, optional
        :param corr_between: correlation matrix (n,n) between input quantities, defaults to None
        :type corr_between: array, optional
        :param return_corr: set to True to return correlation matrix of measurand, defaults to False
        :type return_corr: bool, optional
        :param return_samples: set to True to return generated samples, defaults to False
        :type return_samples: bool, optional
        :param repeat_dims: set to positive integer(s) to select the axis which has repeated measurements. The calculations will be performed seperately for each of the repeated measurments and then combined, in order to save memory and speed up the process.  Defaults to -99, for which there is no reduction in dimensionality..
        :type repeat_dims: integer or list of 2 integers, optional
        :param corr_axis: set to positive integer to select the axis used in the correlation matrix. The correlation matrix will then be averaged over other dimensions. Defaults to -99, for which the input array will be flattened and the full correlation matrix calculated.
        :type corr_axis: integer, optional
        :param fixed_corr_var: set to integer to copy the correlation matrix of the dimiension the integer refers to. Set to True to automatically detect if only one uncertainty is present and the correlation matrix of that dimension should be copied. Defaults to False.
        :type fixed_corr_var: bool or integer, optional
        :param output_vars: number of output parameters in the measurement function. Defaults to 1.
        :type output_vars: integer, optional
        :param PD_corr: set to True to make sure returned correlation matrices are positive semi-definite, default to True
        :type PD_corr: bool, optional
        :return: uncertainties on measurand
        :rtype: array
        """
        (
            yshapes,
            u_x,
            n_repeats,
            repeat_shape,
            repeat_dims,
            corr_axis,
            fixed_corr,
        ) = self.perform_checks(
            func,
            x,
            u_x,
            corr_x,
            repeat_dims,
            corr_axis,
            output_vars,
            fixed_corr_var,
            param_fixed,
        )
        if n_repeats > 0:
            outs = np.empty(n_repeats, dtype=object)
            for i in range(n_repeats):
                xb, u_xb = util.select_repeated_x(
                    x, u_x, param_fixed, i, repeat_dims, repeat_shape
                )
                outs[i] = self.propagate_random(
                    func,
                    xb,
                    u_xb,
                    corr_x,
                    param_fixed,
                    corr_between,
                    return_corr,
                    return_samples,
                    -99,
                    corr_axis=corr_axis,
                    output_vars=output_vars,
                    fixed_corr_var=fixed_corr_var,
                    PD_corr=False,
                )

            return self.combine_repeated_outs(
                outs,
                yshapes,
                len(x),
                n_repeats,
                repeat_shape,
                repeat_dims,
                return_corr,
                return_samples,
                output_vars,
            )

        else:
            MC_data = np.empty(len(x), dtype=np.ndarray)
            for i in range(len(x)):
                if corr_x is None:
                    MC_data[i] = self.generate_samples_random(x[i], u_x[i])
                elif corr_x[i] is None or corr_x[i] == "rand":
                    MC_data[i] = self.generate_samples_random(x[i], u_x[i])
                elif corr_x[i] == "syst":
                    MC_data[i] = self.generate_samples_systematic(x[i], u_x[i])
                else:
                    MC_data[i] = self.generate_samples_correlated(x, u_x, corr_x, i)

            if corr_between is not None:
                MC_data = self.correlate_samples_corr(MC_data, corr_between)

            return self.process_samples(
                func,
                MC_data,
                return_corr,
                return_samples,
                yshapes,
                corr_axis,
                fixed_corr,
                PD_corr,
                output_vars,
            )

    def propagate_systematic(
        self,
        func,
        x,
        u_x,
        corr_x=None,
        param_fixed=None,
        corr_between=None,
        return_corr=False,
        return_samples=False,
        repeat_dims=-99,
        corr_axis=-99,
        fixed_corr_var=False,
        output_vars=1,
        PD_corr=True,
    ):
        """
        Propagate systematic uncertainties through measurement function with n input quantities.
        Input quantities can be floats, vectors (1d-array) or images (2d-array).
        Systematic uncertainties arise when there is full correlation between repeated measurements.
        There is a often also a correlation between measurements along the dimensions that is not one of the repeat_dims.

        :param func: measurement function
        :type func: function
        :param x: list of input quantities (usually numpy arrays)
        :type x: list[array]
        :param u_x: list of systematic uncertainties on input quantities (usually numpy arrays)
        :type u_x: list[array]
        :param corr_x: list of correlation matrices (n,n) along non-repeating axis, defaults to None. Can optionally be set to "rand" (diagonal correlation matrix), "syst" (correlation matrix of ones) or a custom correlation matrix.
        :type corr_x: list[array], optional
        :param param_fixed: when repeat_dims>=0, set to true or false to indicate for each input quantity whether it has repeated measurements that should be split (param_fixed=False) or whether the input is fixed (param fixed=True), defaults to None (no inputs fixed).
        :type param_fixed: list of bools, optional
        :param corr_between: correlation matrix (n,n) between input quantities, defaults to None
        :type corr_between: array, optional
        :param return_corr: set to True to return correlation matrix of measurand, defaults to False
        :type return_corr: bool, optional
        :param return_samples: set to True to return generated samples, defaults to False
        :type return_samples: bool, optional
        :param repeat_dims: set to positive integer(s) to select the axis which has repeated measurements. The calculations will be performed seperately for each of the repeated measurments and then combined, in order to save memory and speed up the process.  Defaults to -99, for which there is no reduction in dimensionality..
        :type repeat_dims: integer or list of 2 integers, optional
        :param corr_axis: set to positive integer to select the axis used in the correlation matrix. The correlation matrix will then be averaged over other dimensions. Defaults to -99, for which the input array will be flattened and the full correlation matrix calculated.
        :type corr_axis: integer, optional
        :param fixed_corr_var: set to integer to copy the correlation matrix of the dimiension the integer refers to. Set to True to automatically detect if only one uncertainty is present and the correlation matrix of that dimension should be copied. Defaults to False.
        :type fixed_corr_var: bool or integer, optional
        :param output_vars: number of output parameters in the measurement function. Defaults to 1.
        :type output_vars: integer, optional
        :param PD_corr: set to True to make sure returned correlation matrices are positive semi-definite, default to True
        :type PD_corr: bool, optional
        :return: uncertainties on measurand
        :rtype: array
        """
        (
            yshapes,
            u_x,
            n_repeats,
            repeat_shape,
            repeat_dims,
            corr_axis,
            fixed_corr,
        ) = self.perform_checks(
            func,
            x,
            u_x,
            corr_x,
            repeat_dims,
            corr_axis,
            output_vars,
            fixed_corr_var,
            param_fixed,
        )

        if n_repeats > 0:
            outs = np.empty(n_repeats, dtype=object)
            for i in range(n_repeats):
                xb, u_xb = util.select_repeated_x(
                    x, u_x, param_fixed, i, repeat_dims, repeat_shape
                )

                outs[i] = self.propagate_systematic(
                    func,
                    xb,
                    u_xb,
                    corr_x,
                    param_fixed,
                    corr_between,
                    return_corr,
                    return_samples,
                    -99,
                    corr_axis=corr_axis,
                    fixed_corr_var=fixed_corr_var,
                    output_vars=output_vars,
                    PD_corr=False,
                )

            return self.combine_repeated_outs(
                outs,
                yshapes,
                len(x),
                n_repeats,
                repeat_shape,
                repeat_dims,
                return_corr,
                return_samples,
                output_vars,
            )

        else:
            MC_data = np.empty(len(x), dtype=np.ndarray)
            for i in range(len(x)):
                if corr_x is None:
                    MC_data[i] = self.generate_samples_systematic(x[i], u_x[i])
                elif corr_x[i] is None or corr_x[i] == "syst":
                    MC_data[i] = self.generate_samples_systematic(x[i], u_x[i])
                elif corr_x[i] == "rand":
                    MC_data[i] = self.generate_samples_random(x[i], u_x[i])
                else:
                    MC_data[i] = self.generate_samples_correlated(x, u_x, corr_x, i)

            if corr_between is not None:
                MC_data = self.correlate_samples_corr(MC_data, corr_between)

            return self.process_samples(
                func,
                MC_data,
                return_corr,
                return_samples,
                yshapes,
                corr_axis,
                fixed_corr,
                PD_corr,
                output_vars,
            )

    def propagate_cov(
        self,
        func,
        x,
        cov_x,
        param_fixed=None,
        corr_between=None,
        return_corr=True,
        return_samples=False,
        repeat_dims=-99,
        corr_axis=-99,
        fixed_corr_var=False,
        output_vars=1,
        PD_corr=True,
    ):
        """
        Propagate uncertainties with given covariance matrix through measurement function with n input quantities.
        Input quantities can be floats, vectors (1d-array) or images (2d-array).
        The covariance matrix can represent the full covariance matrix between all measurements in all dimensions.
        Alternatively if there are repeated measurements specified in repeat_dims, the covariance matrix is given
        for the covariance along the dimension that is not one of the repeat_dims.

        :param func: measurement function
        :type func: function
        :param x: list of input quantities (usually numpy arrays)
        :type x: list[array]
        :param cov_x: list of covariance matrices on input quantities (usually numpy arrays). In case the input quantity is an array of shape (m,o), the covariance matrix needs to be given as an array of shape (m*o,m*o).
        :type cov_x: list[array]
        :param param_fixed: when repeat_dims>=0, set to true or false to indicate for each input quantity whether it has repeated measurements that should be split (param_fixed=False) or whether the input is fixed (param fixed=True), defaults to None (no inputs fixed).
        :type param_fixed: list of bools, optional
        :param corr_between: covariance matrix (n,n) between input quantities, defaults to None
        :type corr_between: array, optional
        :param return_corr: set to True to return correlation matrix of measurand, defaults to True
        :type return_corr: bool, optional
        :param return_samples: set to True to return generated samples, defaults to False
        :type return_samples: bool, optional
        :param repeat_dims: set to positive integer(s) to select the axis which has repeated measurements. The calculations will be performed seperately for each of the repeated measurments and then combined, in order to save memory and speed up the process.  Defaults to -99, for which there is no reduction in dimensionality..
        :type repeat_dims: integer or list of 2 integers, optional
        :param corr_axis: set to positive integer to select the axis used in the correlation matrix. The correlation matrix will then be averaged over other dimensions. Defaults to -99, for which the input array will be flattened and the full correlation matrix calculated.
        :type corr_axis: integer, optional
        :param fixed_corr_var: set to integer to copy the correlation matrix of the dimiension the integer refers to. Set to True to automatically detect if only one uncertainty is present and the correlation matrix of that dimension should be copied. Defaults to False.
        :type fixed_corr_var: bool or integer, optional
        :param output_vars: number of output parameters in the measurement function. Defaults to 1.
        :type output_vars: integer, optional
        :param PD_corr: set to True to make sure returned correlation matrices are positive semi-definite, default to True
        :type PD_corr: bool, optional
        :return: uncertainties on measurand
        :rtype: array
        """
        (
            yshapes,
            u_x,
            n_repeats,
            repeat_shape,
            repeat_dims,
            corr_axis,
            fixed_corr,
        ) = self.perform_checks(
            func,
            x,
            cov_x,
            None,
            repeat_dims,
            corr_axis,
            output_vars,
            fixed_corr_var,
            param_fixed,
        )

        if n_repeats > 0:
            outs = np.empty(n_repeats, dtype=object)
            for i in range(n_repeats):
                xb, _ = util.select_repeated_x(
                    x, x, param_fixed, i, repeat_dims, repeat_shape
                )
                outs[i] = self.propagate_cov(
                    func,
                    xb,
                    cov_x,
                    param_fixed,
                    corr_between,
                    return_corr,
                    return_samples,
                    -99,
                    corr_axis=corr_axis,
                    output_vars=output_vars,
                    PD_corr=False,
                )

            return self.combine_repeated_outs(
                outs,
                yshapes,
                len(x),
                n_repeats,
                repeat_shape,
                repeat_dims,
                return_corr,
                return_samples,
                output_vars,
            )

        else:
            MC_data = np.empty(len(x), dtype=np.ndarray)
            for i in range(len(x)):
                if not hasattr(x[i], "__len__"):
                    MC_data[i] = self.generate_samples_systematic(x[i], cov_x[i])
                elif all(
                    (cov_x[i] == 0).flatten()
                ):  # This is the case if one of the variables has no uncertainty
                    MC_data[i] = np.tile(x[i].flatten(), (self.MCsteps, 1)).T
                elif param_fixed is not None:
                    if param_fixed[i] and (len(x[i].shape) == 2):
                        MC_data[i] = np.array(
                            [
                                self.generate_samples_cov(
                                    x[i][:, j].flatten(), cov_x[i]
                                ).reshape(x[i][:, j].shape + (self.MCsteps,))
                                for j in range(x[i].shape[1])
                            ]
                        ).T
                        MC_data[i] = np.moveaxis(MC_data[i], 0, 1)
                    else:
                        MC_data[i] = self.generate_samples_cov(
                            x[i].flatten(), cov_x[i]
                        ).reshape(x[i].shape + (self.MCsteps,))
                else:
                    MC_data[i] = self.generate_samples_cov(
                        x[i].flatten(), cov_x[i]
                    ).reshape(x[i].shape + (self.MCsteps,))

            if corr_between is not None:
                MC_data = self.correlate_samples_corr(MC_data, corr_between)

        return self.process_samples(
            func,
            MC_data,
            return_corr,
            return_samples,
            yshapes,
            corr_axis,
            fixed_corr,
            PD_corr,
            output_vars,
        )

    def perform_checks(
        self,
        func,
        x,
        u_x,
        corr_x,
        repeat_dims,
        corr_axis,
        output_vars,
        fixed_corr_var,
        param_fixed,
    ):
        """
        Perform checks on the input parameters and set up the appropriate keywords for further processing

        :param func: measurement function
        :type func: function
        :param x: list of input quantities (usually numpy arrays)
        :type x: list[array]
        :param u_x: list of uncertainties/covariances on input quantities (usually numpy arrays)
        :type u_x: list[array]
        :param corr_x: list of correlation matrices (n,n) along non-repeating axis, defaults to None. Can optionally be set to "rand" (diagonal correlation matrix), "syst" (correlation matrix of ones) or a custom correlation matrix.
        :type corr_x: list[array], optional
        :param repeat_dims: set to positive integer(s) to select the axis which has repeated measurements. The calculations will be performed seperately for each of the repeated measurments and then combined, in order to save memory and speed up the process.  Defaults to -99, for which there is no reduction in dimensionality..
        :type repeat_dims: integer or list of 2 integers, optional
        :param corr_axis: set to positive integer to select the axis used in the correlation matrix. The correlation matrix will then be averaged over other dimensions. Defaults to -99, for which the input array will be flattened and the full correlation matrix calculated.
        :type corr_axis: integer, optional
        :param output_vars: number of output parameters in the measurement function. Defaults to 1.
        :type output_vars: integer, optional
        :param fixed_corr_var: set to integer to copy the correlation matrix of the dimiension the integer refers to. Set to True to automatically detect if only one uncertainty is present and the correlation matrix of that dimension should be copied. Defaults to False.
        :type fixed_corr_var: bool or integer, optional
        :param param_fixed: when repeat_dims>=0, set to true or false to indicate for each input quantity whether it has repeated measurements that should be split (param_fixed=False) or whether the input is fixed (param fixed=True), defaults to None (no inputs fixed).
        :type param_fixed: list of bools, optional
        :return: yshape,u_x,repeat_axis,repeat_dims,corr_axis,fixed_corr
        :rtype: tuple, list[array], int, int, int, array
        """

        # find the shape
        if output_vars == 1:
            yshape = np.array(func(*x)).shape
            yshapes = [np.array(func(*x)).shape]
        else:
            yshape = np.array(func(*x)[0]).shape
            yshapes = [np.array(func(*x)[i]).shape for i in range(output_vars)]

        shapewarning = False
        for i in range(len(x)):
            if hasattr(x[i], "__len__"):
                if param_fixed is not None:
                    if (
                        x[i].shape != yshape
                        and self.parallel_cores == 0
                        and not param_fixed[i]
                    ):
                        shapewarning = True
                else:
                    if x[i].shape != yshape and self.parallel_cores == 0:
                        shapewarning = True
            elif self.parallel_cores == 0:
                shapewarning = True

        if shapewarning:
            warnings.warn(
                "It looks like one of your input quantities is not an array or does not "
                "have the same shape as the measurand. This is not a problem, but means "
                "you likely cannot use array operations in your measurement function. "
                "You might need to set parallel_cores to 1 or higher when creating "
                "your MCPropagation object."
            )

        # Check for which input quantities there is no uncertainty,
        # replacing Nones with zeros where necessary.
        # Count the number of non-zero uncertainties. If this number is one, the
        # correlation matrix for the measurand will be the same as for this input qty.

        count = 0
        for i in range(len(x)):
            if u_x[i] is None:
                if hasattr(x[i], "__len__"):
                    u_x[i] = np.zeros(x[i].shape)
                else:
                    u_x[i] = 0.0
            if np.sum(u_x[i]) != 0 and fixed_corr_var == True:
                count += 1
                var = i
            if corr_x is not None:
                if corr_x[i] is not None:
                    if not isinstance(corr_x[i], str):
                        if np.any(corr_x[i] > 1.000001):
                            raise ValueError(
                                "One of the provided correlation matrices "
                                "has elements >1."
                            )

        if count == 1:
            fixed_corr_var = var
        else:
            fixed_corr_var = -99

        if fixed_corr_var >= 0 and corr_x is not None:
            if corr_x[fixed_corr_var] == "rand":
                fixed_corr = np.eye(len(u_x[fixed_corr_var]))
            elif corr_x[fixed_corr_var] == "syst":
                fixed_corr = np.ones(
                    (len(u_x[fixed_corr_var]), len(u_x[fixed_corr_var]))
                )
            else:
                fixed_corr = corr_x[fixed_corr_var]

        else:
            fixed_corr = None

        # Set up repeat_axis and repeat_dims for proper use in recursive function.
        if isinstance(repeat_dims, int):
            repeat_dims = [repeat_dims]

        if len(repeat_dims) == 1:
            if repeat_dims[0] >= 0:
                n_repeats = yshape[repeat_dims[0]]
                if corr_axis > repeat_dims[0]:
                    corr_axis -= 1
                elif corr_axis == repeat_dims[0]:
                    print("corr_axis and repeat_axis keywords should not be the same.")
                    exit()
            else:
                n_repeats = 0
            repeat_shape = (n_repeats,)  # repeat_dims = -99
        else:
            repeat_dims = -np.sort(-np.array(repeat_dims))
            n_repeats = yshape[repeat_dims[0]] * yshape[repeat_dims[1]]
            repeat_shape = (yshape[repeat_dims[0]], yshape[repeat_dims[1]])
            if corr_axis > repeat_dims[0]:
                corr_axis -= 1
            elif corr_axis == repeat_dims[0]:
                print("corr_axis and repeat_axis keywords should not be the same.")
                exit()
            if corr_axis > repeat_dims[1]:
                corr_axis -= 1
            elif corr_axis == repeat_dims[1]:
                print("corr_axis and repeat_axis keywords should not be the same.")
                exit()

        return yshapes, u_x, n_repeats, repeat_shape, repeat_dims, corr_axis, fixed_corr

    def combine_repeated_outs(
        self,
        outs,
        yshapes,
        lenx,
        n_repeats,
        repeat_shape,
        repeat_dims,
        return_corr,
        return_samples,
        output_vars,
    ):
        """
        Combine the outputs of the repeated measurements into one results array

        :param outs: list of outputs of the repeated measurements
        :type outs: list[array]
        :param yshapes: shape of the measurand
        :type yshapes: tuple
        :param lenx: number of input quantities
        :type lenx: int
        :param n_repeats: number of repeated measurements
        :type n_repeats: int
        :param repeat_shape: shape along which the measurements are repeated
        :type repeat_shape: tuple
        :param repeat_dims: set to positive integer(s) to select the axis which has repeated measurements. The calculations will be performed seperately for each of the repeated measurments and then combined, in order to save memory and speed up the process.  Defaults to -99, for which there is no reduction in dimensionality..
        :type repeat_dims: integer or list of 2 integers, optional
        :param return_corr: set to True to return correlation matrix of measurand, defaults to True
        :type return_corr: bool, optional
        :param return_Jacobian: set to True to return Jacobian, defaults to False
        :type return_Jacobian: bool, optional
        :param output_vars: number of output parameters in the measurement function. Defaults to 1.
        :type output_vars: integer, optional
        :return: combined outputs
        :rtype: list[array]
        """
        if not return_corr and output_vars == 1 and not return_samples:
            u_func = np.array([outs[i] for i in range(n_repeats)])

        elif output_vars == 1:
            u_func = np.array([outs[i][0] for i in range(n_repeats)])

        elif output_vars > 1 and not return_corr and not return_samples:
            u_func = np.array(
                [[outs[i][ii] for i in range(n_repeats)] for ii in range(output_vars)]
            )

        else:
            u_func = np.array(
                [
                    [outs[i][0][ii] for i in range(n_repeats)]
                    for ii in range(output_vars)
                ]
            )

        if len(repeat_dims) == 1:
            if output_vars == 1:
                u_func = np.moveaxis(u_func, 0, repeat_dims[0])
            else:
                if all([yshapes[i]==yshapes[0] for i in range(len(yshapes))]):
                    u_func = np.moveaxis(u_func, 1, repeat_dims[0] + 1)
                else:
                    u_funcb=u_func[:]
                    u_func=np.empty(output_vars,dtype=object)
                    for i in range(output_vars):
                        u_func[i]=np.empty(yshapes[i])
                        if len(yshapes[i]) == 0:
                            u_func[i]= u_funcb[i]
                        elif len(yshapes[i]) == 1:
                            for ii in range(yshapes[i][0]):
                                u_func[i][ii] = u_funcb[i][ii]
                        elif len(yshapes[i]) == 2:
                            for ii in range(yshapes[i][0]):
                                for iii in range(yshapes[i][1]):
                                        u_func[i][ii,iii] = u_funcb[i][iii][ii]
                        else:
                            print("this shape is not supported")

        else:
            if output_vars == 1:
                u_func = u_func.reshape(repeat_shape + (-1,))
                u_func = np.moveaxis(u_func, 0, repeat_dims[0])
                u_func = np.moveaxis(u_func, 0, repeat_dims[1])
            else:
                u_func = u_func.reshape((output_vars,) + repeat_shape + (-1,))
                u_func = np.moveaxis(u_func, 1, repeat_dims[0] + 1)
                u_func = np.moveaxis(u_func, 1, repeat_dims[1] + 1)

        if (output_vars == 1 and u_func.shape != yshapes[0]) or (
            output_vars > 1 and u_func[0].shape != yshapes[0]
        ):
            print(u_func.shape, yshapes[0])
            raise ValueError(
                "The shape of the uncertainties does not match the shape"
                "of the measurand. This is likely a problem with combining"
                "repeated measurements (repeat_dims keyword)."
            )

        if not return_corr and not return_samples:
            return u_func

        else:
            returns = np.empty(len(outs[0]), dtype=object)
            returns[0] = u_func
            extra_index = 0
            if return_corr:
                corr = np.mean([outs[i][1] for i in range(n_repeats)], axis=0)
                if all([yshapes[i]==yshapes[0] for i in range(len(yshapes))]):
                    if output_vars > 1:
                        for j in range(output_vars):
                            if not util.isPD(corr[j]):
                                corr[j] = util.nearestPD_cholesky(
                                    corr[j], corr=True, return_cholesky=False
                                )
                    else:
                        if not util.isPD(corr):
                            corr = util.nearestPD_cholesky(
                                corr, corr=True, return_cholesky=False
                            )

                returns[1] = corr
                extra_index += 1

            if output_vars > 1:
                if all([yshapes[i]==yshapes[0] for i in range(len(yshapes))]):
                    corr_out = np.mean(
                        [outs[i][1 + extra_index] for i in range(n_repeats)], axis=0
                    )
                    if not util.isPD(corr_out):
                        corr_out = util.nearestPD_cholesky(
                            corr_out, corr=True, return_cholesky=False
                        )
                else:
                    corr_out=None
                returns[1 + extra_index] = corr_out
                extra_index += 1

            if return_samples:
                returns[1 + extra_index] = [
                    np.vstack([outs[i][1 + extra_index][k] for i in range(n_repeats)])
                    for k in range(lenx)
                ]
                returns[2 + extra_index] = [
                    np.vstack([outs[i][2 + extra_index][k] for i in range(n_repeats)])
                    for k in range(lenx)
                ]

            return returns

    def process_samples(
        self,
        func,
        data,
        return_corr,
        return_samples,
        yshapes,
        corr_axis=-99,
        fixed_corr=None,
        PD_corr=True,
        output_vars=1,
    ):
        """
        Run the MC-generated samples of input quantities through the measurement function and calculate
        correlation matrix if required.

        :param func: measurement function
        :type func: function
        :param data: MC-generated samples of input quantities
        :type data: array[array]
        :param return_corr: set to True to return correlation matrix of measurand
        :type return_corr: bool
        :param return_samples: set to True to return generated samples
        :type return_samples: bool
        :param corr_axis: set to positive integer to select the axis used in the correlation matrix. The correlation matrix will then be averaged over other dimensions. Defaults to -99, for which the input array will be flattened and the full correlation matrix calculated.
        :type corr_axis: integer, optional
        :param fixed_corr: correlation matrix to be copied without changing, defaults to None (correlation matrix is calculated rather than copied)
        :type fixed_corr: array
        :param PD_corr: set to True to make sure returned correlation matrices are positive semi-definite, default to True
        :type PD_corr: bool, optional
        :param output_vars: number of output parameters in the measurement function. Defaults to 1.
        :type output_vars: integer, optional
        :return: uncertainties on measurand
        :rtype: array
        """
        if self.parallel_cores == 0:
            MC_y = np.array(func(*data))

        elif self.parallel_cores == 1:
            # In order to Process the MC iterations separately, the array with the input quantities has to be reordered
            # so that it has the same length (i.e. the first dimension) as the number of MCsteps.
            # First we move the axis with the same length as self.MCsteps from the last dimension to the fist dimension
            data2 = [np.moveaxis(dat, -1, 0) for dat in data]
            if output_vars==1 or all([yshapes[i]==yshapes[0] for i in range(len(yshapes))]):
                # The function can then be applied to each of these MCsteps
                MC_y2 = np.array(list(map(func,*data2)))

                # We then reorder to bring it back to the original shape
                MC_y = np.moveaxis(MC_y2, 0, -1)

            else:
                MC_y = np.empty(output_vars,dtype=object)
                MC_y2 = np.array(list(map(func,*data2)))

                for i in range(output_vars):
                    MC_y[i]=np.empty(yshapes[i]+(self.MCsteps,))
                    if len(yshapes[i])==0:
                        for j in range(self.MCsteps):
                            MC_y[i][j]=MC_y2[j,i]
                    elif len(yshapes[i])==1:
                        for ii in range(yshapes[i][0]):
                            for j in range(self.MCsteps):
                                MC_y[i][ii,j] =MC_y2[j,i][ii]
                    elif len(yshapes[i])==2:
                        for ii in range(yshapes[i][0]):
                            for iii in range(yshapes[i][1]):
                                for j in range(self.MCsteps):
                                    MC_y[i][ii,iii,j] =MC_y2[j,i][ii,iii]
                    else:
                        print("this shape is not supported")

        else:
            # We again need to reorder the input quantities samples in order to be able to pass them to p.starmap
            # We here use lists to iterate over and order them slightly different as the case above.
            data2 = [
                [data[j][..., i] for j in range(len(data))] for i in range(self.MCsteps)
            ]
            pool = Pool(self.parallel_cores)
            MC_y2 = np.array(pool.starmap(func, data2))
            pool.close()
            del pool
            if output_vars == 1:
                MC_y = np.moveaxis(MC_y2,0,-1)
            else:
                MC_y = np.empty(output_vars,dtype=object)
                for i in range(output_vars):
                    MC_y[i] = np.moveaxis(MC_y2[i],0,-1)

        # if hasattr(MC_y[0,0], '__len__'):
        #     print(yshape,np.array(MC_y[0,0]).shape,np.array(MC_y[1,0]).shape,np.array(MC_y[2,0]).shape,np.array(MC_y[3,0]).shape)
        if output_vars == 1:
            u_func = np.std(MC_y, axis=-1)
        else:
            u_func = np.empty(output_vars,dtype=object)
            for i in range(output_vars):
                u_func[i] = np.std(np.array(MC_y[i]), axis=-1)
        if not return_corr:
            if return_samples:
                return u_func, MC_y, data
            else:
                return u_func
        else:
            if output_vars == 1:
                if fixed_corr is None:
                    corr_y = self.calculate_corr(MC_y, corr_axis).astype("float32")
                    if PD_corr:
                        if not util.isPD(corr_y):
                            corr_y = util.nearestPD_cholesky(
                                corr_y, corr=True, return_cholesky=False
                            )
                else:
                    corr_y = fixed_corr
                if return_samples:
                    return u_func, corr_y, MC_y, data
                else:
                    return u_func, corr_y

            else:
                # create an empty arrays and then populate it with the correlation matrix for each output parameter individually
                corr_ys = np.empty(output_vars, dtype=object)
                for i in range(output_vars):
                    if fixed_corr is None:
                        corr_ys[i] = self.calculate_corr(MC_y[i], corr_axis).astype(
                            "float32"
                        )
                        if PD_corr:
                            if not util.isPD(corr_ys[i]):
                                corr_ys[i] = util.nearestPD_cholesky(
                                    corr_ys[i], corr=True, return_cholesky=False
                                )
                    else:
                        corr_ys[i] = fixed_corr
                # calculate correlation matrix between the different outputs produced by the measurement function.
                if all([yshapes[i]==yshapes[0] for i in range(len(yshapes))]):
                    corr_out = np.corrcoef(MC_y.reshape((output_vars, -1))).astype(
                        "float32"
                    )
                else:
                    corr_out = None# np.corrcoef([[np.mean(MC_y[i][:,j]) for j in range(self.MCsteps)] for i in range(output_vars)]).astype(
                       # "float32")

                if PD_corr:
                    if not util.isPD(corr_out):
                        corr_out = util.nearestPD_cholesky(
                            corr_out, corr=True, return_cholesky=False
                        )
                if return_samples:
                    return u_func, corr_ys, corr_out, MC_y, data
                else:
                    return u_func, corr_ys, corr_out

    def calculate_corr(self, MC_y, corr_axis=-99):
        """
        Calculate the correlation matrix between the MC-generated samples of output quantities.
        If corr_axis is specified, this axis will be the one used to calculate the correlation matrix (e.g. if corr_axis=0 and x.shape[0]=n, the correlation matrix will have shape (n,n)).
        This will be done for each combination of parameters in the other dimensions and the resulting correlation matrices are averaged.

        :param MC_y: MC-generated samples of the output quantities (measurands)
        :type MC_y: array
        :param corr_axis: set to positive integer to select the axis used in the correlation matrix. The correlation matrix will then be averaged over other dimensions. Defaults to -99, for which the input array will be flattened and the full correlation matrix calculated.
        :type corr_axis: integer, optional
        :return: correlation matrix
        :rtype: array
        """
        # print("the shape is:",MC_y.shape)

        if len(MC_y.shape) < 3:
            corr_y = np.corrcoef(MC_y)

        elif len(MC_y.shape) == 3:
            if corr_axis == 0:
                corr_ys = np.empty(len(MC_y[0]), dtype=object)
                for i in range(len(MC_y[0])):
                    corr_ys[i] = np.corrcoef(MC_y[:, i])
                corr_y = np.mean(corr_ys, axis=0)

            elif corr_axis == 1:
                corr_ys = np.empty(len(MC_y), dtype=object)
                for i in range(len(MC_y)):
                    corr_ys[i] = np.corrcoef(MC_y[i])
                corr_y = np.mean(corr_ys, axis=0)

            else:
                MC_y = MC_y.reshape((MC_y.shape[0] * MC_y.shape[1], self.MCsteps))
                corr_y = np.corrcoef(MC_y)

        elif len(MC_y.shape) == 4:
            if corr_axis == 0:
                corr_ys = np.empty(len(MC_y[0]) * len(MC_y[0, 0]), dtype=object)
                for i in range(len(MC_y[0])):
                    for j in range(len(MC_y[0, 0])):
                        corr_ys[i + j * len(MC_y[0])] = np.corrcoef(MC_y[:, i, j])
                corr_y = np.mean(corr_ys, axis=0)

            elif corr_axis == 1:
                corr_ys = np.empty(len(MC_y) * len(MC_y[0, 0]), dtype=object)
                for i in range(len(MC_y)):
                    for j in range(len(MC_y[0, 0])):
                        corr_ys[i + j * len(MC_y)] = np.corrcoef(MC_y[i, :, j])
                corr_y = np.mean(corr_ys, axis=0)

            elif corr_axis == 2:
                corr_ys = np.empty(len(MC_y) * len(MC_y[0]), dtype=object)
                for i in range(len(MC_y)):
                    for j in range(len(MC_y[0])):
                        corr_ys[i + j * len(MC_y)] = np.corrcoef(MC_y[i, j])
                corr_y = np.mean(corr_ys, axis=0)
            else:
                MC_y = MC_y.reshape(
                    (MC_y.shape[0] * MC_y.shape[1] * MC_y.shape[2], self.MCsteps)
                )
                corr_y = np.corrcoef(MC_y)
        else:
            print(
                "MC_y has too high dimensions. Reduce the dimensionality of the input data"
            )
            exit()

        return corr_y

    def generate_samples_correlated(self, x, u_x, corr_x, i):
        """
        Generate correlated MC samples of input quantity with given uncertainties and correlation matrix.
        Samples are generated using generate_samples_cov() after matching up the uncertainties to the right correlation matrix.
        It is possible to provide one correlation matrix to be used for each measurement (which each have an uncertainty) or a correlation matrix per measurement.

        :param x: list of input quantities (usually numpy arrays)
        :type x: list[array]
        :param u_x: list of uncertainties/covariances on input quantities (usually numpy arrays)
        :type u_x: list[array]
        :param corr_x: list of correlation matrices (n,n) along non-repeating axis, or list of correlation matrices for each repeated measurement.
        :type corr_x: list[array], optional
        :param i: index of the input quantity (in x)
        :type i: int
        :return: generated samples
        :rtype: array
        """
        if len(x[i].shape) == 2:
            if len(corr_x[i]) == len(u_x[i]):
                MC_data = np.zeros((u_x[i].shape) + (self.MCsteps,))
                for j in range(len(u_x[i][0])):
                    cov_x = util.convert_corr_to_cov(corr_x[i], u_x[i][:, j])
                    MC_data[:, j, :] = self.generate_samples_cov(
                        x[i][:, j].flatten(), cov_x
                    ).reshape(x[i][:, j].shape + (self.MCsteps,))
            else:
                MC_data = np.zeros((u_x[i].shape) + (self.MCsteps,))
                for j in range(len(u_x[i][:, 0])):
                    cov_x = util.convert_corr_to_cov(corr_x[i], u_x[i][j])
                    MC_data[j, :, :] = self.generate_samples_cov(
                        x[i][j].flatten(), cov_x
                    ).reshape(x[i][j].shape + (self.MCsteps,))
        else:
            cov_x = util.convert_corr_to_cov(corr_x[i], u_x[i])
            MC_data = self.generate_samples_cov(x[i].flatten(), cov_x).reshape(
                x[i].shape + (self.MCsteps,)
            )

        return MC_data

    def generate_samples_random(self, param, u_param):
        """
        Generate MC samples of input quantity with random (Gaussian) uncertainties.

        :param param: values of input quantity (mean of distribution)
        :type param: float or array
        :param u_param: uncertainties on input quantity (std of distribution)
        :type u_param: float or array
        :return: generated samples
        :rtype: array
        """
        if not hasattr(param, "__len__"):
            return np.random.normal(size=self.MCsteps) * u_param + param
        elif len(param.shape) == 1:
            return (
                np.random.normal(size=(len(param), self.MCsteps)) * u_param[:, None]
                + param[:, None]
            )
        elif len(param.shape) == 2:
            return (
                np.random.normal(size=param.shape + (self.MCsteps,))
                * u_param[:, :, None]
                + param[:, :, None]
            )
        elif len(param.shape) == 3:
            return (
                np.random.normal(size=param.shape + (self.MCsteps,))
                * u_param[:, :, :, None]
                + param[:, :, :, None]
            )
        else:
            print("parameter shape not supported")
            exit()

    def generate_samples_systematic(self, param, u_param):
        """
        Generate correlated MC samples of input quantity with systematic (Gaussian) uncertainties.

        :param param: values of input quantity (mean of distribution)
        :type param: float or array
        :param u_param: uncertainties on input quantity (std of distribution)
        :type u_param: float or array
        :return: generated samples
        :rtype: array
        """
        if not hasattr(param, "__len__"):
            return np.random.normal(size=self.MCsteps) * u_param + param
        elif len(param.shape) == 1:
            return (
                np.dot(u_param[:, None], np.random.normal(size=self.MCsteps)[None, :])
                + param[:, None]
            )
        elif len(param.shape) == 2:
            return (
                np.dot(
                    u_param[:, :, None],
                    np.random.normal(size=self.MCsteps)[:, None, None],
                )[:, :, :, 0]
                + param[:, :, None]
            )
        elif len(param.shape) == 3:
            return (
                np.dot(
                    u_param[:, :, :, None],
                    np.random.normal(size=self.MCsteps)[:, None, None, None],
                )[:, :, :, :, 0, 0]
                + param[:, :, :, None]
            )
        else:
            print("parameter shape not supported")
            exit()

    def generate_samples_cov(self, param, cov_param):
        """
        Generate correlated MC samples of input quantity with a given covariance matrix.
        Samples are generated independent and then correlated using Cholesky decomposition.

        :param param: values of input quantity (mean of distribution)
        :type param: array
        :param cov_param: covariance matrix for input quantity
        :type cov_param: array
        :return: generated samples
        :rtype: array
        """
        try:
            L = np.linalg.cholesky(cov_param)
        except:
            L = util.nearestPD_cholesky(cov_param)

        return (
            np.dot(L, np.random.normal(size=(len(param), self.MCsteps)))
            + param[:, None]
        )

    def correlate_samples_corr(self, samples, corr):
        """
        Method to correlate independent samples of input quantities using correlation matrix and Cholesky decomposition.

        :param samples: independent samples of input quantities
        :type samples: array[array]
        :param corr: correlation matrix between input quantities
        :type corr: array
        :return: correlated samples of input quantities
        :rtype: array[array]
        """
        if np.max(corr) > 1.000001 or len(corr) != len(samples):
            raise ValueError(
                "The correlation matrix between variables is not the right shape or has elements >1."
            )
        else:
            try:
                L = np.array(np.linalg.cholesky(corr))
            except:
                L = util.nearestPD_cholesky(corr)

            # Cholesky needs to be applied to Gaussian distributions with mean=0 and std=1,
            # We first calculate the mean and std for each input quantity
            means = np.array([np.mean(samples[i]) for i in range(len(samples))])
            stds = np.array([np.std(samples[i]) for i in range(len(samples))])

            # We normalise the samples with the mean and std, then apply Cholesky, and finally reapply the mean and std.
            if all(stds != 0):
                return np.dot(L, (samples - means) / stds) * stds + means

            # If any of the variables has no uncertainty, the normalisation will fail. Instead we leave the parameters without uncertainty unchanged.
            else:
                samples_out = samples[:]
                id_nonzero = np.where(stds != 0)
                samples_out[id_nonzero] = (
                    np.dot(
                        L[id_nonzero][:, id_nonzero],
                        (samples[id_nonzero] - means[id_nonzero]) / stds[id_nonzero],
                    )[:, 0]
                    * stds[id_nonzero]
                    + means[id_nonzero]
                )
                return samples_out



from punpy.mc.mc_propagation import MCPropagation
from punpy.lpu.lpu_propagation import LPUPropagation
from punpy.mc.MCMC_retrieval import MCMCRetrieval
from punpy.lpu.lpu_retrieval import LPURetrieval
from punpy.utilities.utilities import (
    calculate_Jacobian,
    calculate_flattened_corr,
    separate_flattened_corr,
    convert_corr_to_cov,
    convert_cov_to_corr,
    correlation_from_covariance,
    uncertainty_from_covariance,
    nearestPD_cholesky,
)
from ._version import get_versions

"""___Authorship___"""
__author__ = "Pieter De Vis"
__created__ = "01/03/2020"
__maintainer__ = "Pieter De Vis"
__email__ = "pieter.de.vis@npl.co.uk"
__status__ = "Development"

__version__ = get_versions()["version"]
del get_versions



"""Run MCMC for retrieval"""

"""___Built-In Modules___"""

"""___Third-Party Modules___"""
import os
from multiprocessing import Pool

import emcee
import numpy as np

os.environ["OMP_NUM_THREADS"] = "1"

"""___NPL Modules___"""

"""___Authorship___"""
__author__ = "Pieter De Vis"
__created__ = "01/03/2020"
__maintainer__ = "Pieter De Vis"
__email__ = "pieter.de.vis@npl.co.uk"
__status__ = "Development"

inds_cache = {}


# lock = threading.Lock()
class MCMCRetrieval:
    def __init__(
        self,
        measurement_function,
        observed,
        syst_uncertainty=None,
        rand_uncertainty=None,
        cov=None,
        parallel_cores=4,
        uplims=+np.inf,
        downlims=-np.inf,
    ):
        self.measurement_function = measurement_function
        self.observed = observed
        self.rand_uncertainty = np.array([rand_uncertainty])
        self.syst_uncertainty = np.array([syst_uncertainty])
        if cov is None:
            self.invcov = cov
        else:
            self.invcov = np.linalg.inv(np.ascontiguousarray(cov))
        self.uplims = np.array(uplims)
        self.downlims = np.array(downlims)
        self.parallel_cores = parallel_cores

    def run_retrieval(
        self, theta_0, nwalkers, steps, burn_in, return_samples=True, return_corr=False
    ):
        # if self.syst_uncertainty is not None:
        #     theta_0=np.append([0.],theta_0)
        ndimw = len(theta_0)
        pos = [
            theta_0 * np.random.normal(1.0, 0.1, theta_0.shape)
            + np.random.normal(0.0, 0.001, theta_0.shape)
            for i in range(nwalkers)
        ]
        # print(self.measurement_function(theta_0))
        if self.parallel_cores > 1:
            p = Pool(self.parallel_cores)
        else:
            p = None
        sampler = emcee.EnsembleSampler(nwalkers, ndimw, self.lnprob, pool=p)
        sampler.run_mcmc(pos, steps, progress=False)

        samples = sampler.chain[:, :, :].reshape((-1, ndimw))[burn_in::]
        medians = np.median(samples, axis=0)
        unc_up = np.percentile(samples, 84, axis=0) - medians
        unc_down = -(np.percentile(samples, 16, axis=0) - medians)
        unc_avg = (unc_up + unc_down) / 2.0
        corr = np.corrcoef(samples.T)

        if return_samples:
            if return_corr:
                return medians, unc_avg, corr, samples
            else:
                return medians, unc_avg, samples

        else:
            if return_corr:
                return medians, unc_avg, corr
            else:
                return medians, unc_avg

    def find_chisum(self, theta):
        model = self.measurement_function(*theta)
        diff = model - self.observed
        if np.isfinite(np.sum(diff)):
            if self.invcov is None:
                return np.sum((diff) ** 2 / self.rand_uncertainty ** 2)
            else:
                # print(diff,np.linalg.inv(self.cov),np.dot(np.dot(diff.T,self.invcov),diff))
                return np.dot(np.dot(diff.T, self.invcov), diff)
        else:
            return np.inf

    def lnlike(self, theta):
        # print(theta,self.find_chisum(theta))
        return -0.5 * (self.find_chisum(theta))

    def lnprior(self, theta):
        if all(self.downlims < theta[1::]) and all(self.uplims > theta[1::]):
            # if self.syst_uncertainty[0] is None:
            return 0
        # else:
        #     return -0.5*(theta[0]**2/self.syst_uncertainty**2)
        else:
            return -np.inf

    def lnprob(self, theta):
        lp_prior = self.lnprior(theta)
        if not np.isfinite(lp_prior):
            return -np.inf
        lp = self.lnlike(theta)
        # if lp < -99999999999:
        #     return -np.inf
        return lp_prior + lp



"""Use Monte Carlo to propagate uncertainties"""

import warnings

import numdifftools as nd
import numpy as np

"""___Authorship___"""
__author__ = "Pieter De Vis"
__created__ = "30/03/2019"
__maintainer__ = "Pieter De Vis"
__email__ = "pieter.de.vis@npl.co.uk"
__status__ = "Development"


def calculate_Jacobian(fun, x, Jx_diag=False, step=None):
    """
    Calculate the local Jacobian of function y=f(x) for a given value of x

    :param fun: flattened measurement function
    :type fun: function
    :param x: flattened local values of input quantities
    :type x: array
    :param Jx_diag: Bool to indicate whether the Jacobian matrix can be described with semi-diagonal elements. With this we mean that the measurand has the same shape as each of the input quantities and the square jacobain between the measurand and each of the input quantities individually, only has diagonal elements. Defaults to False
    :rtype Jx_diag: bool, optional
    :return: Jacobian
    :rtype: array
    """
    Jfun = nd.Jacobian(fun, step=step)

    if Jx_diag:
        y = fun(x)
        Jfun = nd.Jacobian(fun)
        Jx = np.zeros((len(x), len(y)))
        print(Jx.shape)
        for j in range(len(y)):
            xj = np.zeros(int(len(x) / len(y)))
            for i in range(len(xj)):
                xj[i] = x[i * len(y) + j]
            print(xj.shape, xj)
            Jxj = Jfun(xj)
            for i in range(len(xj)):
                Jx[i * len(y) + j, j] = Jxj[0][i]
    else:
        Jx = Jfun(x)

    if len(Jx) != len(fun(x).flatten()):
        warnings.warn(
            "Dimensions of the Jacobian were flipped because its shape "
            "didn't match the shape of the output of the function "
            "(probably because there was only 1 input qty)."
        )
        Jx = Jx.T

    return Jx


def calculate_flattened_corr(corrs, corr_between):
    """
    Combine correlation matrices for different input quantities, with a correlation
    matrix that gives the correlation between the input quantities into a full
    (flattened) correlation matrix combining the two.

    :param corrs: list of correlation matrices for each input quantity
    :type corrs: list[array]
    :param corr_between: correlation matrix between the input quantities
    :type corr_between: array
    :return: full correlation matrix combining the correlation matrices
    :rtype: array
    """
    totcorrlen = 0
    for i in range(len(corrs)):
        totcorrlen += len(corrs[i])
    totcorr = np.eye(totcorrlen)
    for i in range(len(corrs)):
        for j in range(len(corrs)):
            if corr_between[i, j] > 0:
                ist = i * len(corrs[i])
                iend = (i + 1) * len(corrs[i])
                jst = j * len(corrs[j])
                jend = (j + 1) * len(corrs[j])
                totcorr[ist:iend, jst:jend] = (
                    corr_between[i, j] * corrs[i] ** 0.5 * corrs[j] ** 0.5
                )
    return totcorr


def separate_flattened_corr(corr, ndim):
    """
    Separate a full (flattened) correlation matrix into a list of correlation matrices
    for each output variable and a correlation matrix between the output variables.

    :param corr: full correlation matrix
    :type corr: array
    :param ndim: number of output variables
    :type ndim: int
    :return: list of correlation matrices for each output variable, correlation matrix between the output variables
    :type corrs: list[array]
    :rtype: list[array], array
    """

    corrs = np.empty(ndim, dtype=object)
    for i in range(ndim):
        corrs[i] = correlation_from_covariance(
            corr[
                int(i * len(corr) / ndim) : int((i + 1) * len(corr) / ndim),
                int(i * len(corr) / ndim) : int((i + 1) * len(corr) / ndim),
            ]
        )

    corrs_between = np.empty((ndim, ndim))
    for i in range(ndim):
        for j in range(ndim):
            corrs_between[i, j] = np.nanmean(
                corr[
                    int(i * len(corr) / ndim) : int((i + 1) * len(corr) / ndim),
                    int(j * len(corr) / ndim) : int((j + 1) * len(corr) / ndim),
                ]
                / corrs[i] ** 0.5
                / corrs[j] ** 0.5
            )

    return corrs, corrs_between


def select_repeated_x(x, u_x, param_fixed, i, repeat_dims, repeat_shape):
    """
    Select one (index i) of multiple repeated entries and return the input quantities and uncertainties for that entry.

    :param x: list of input quantities (usually numpy arrays)
    :type x: list[array]
    :param u_x: list of uncertainties/covariances on input quantities (usually numpy arrays)
    :type u_x: list[array]
    :param param_fixed: when repeat_dims>=0, set to true or false to indicate for each input quantity whether it has repeated measurements that should be split (param_fixed=False) or whether the input is fixed (param fixed=True), defaults to None (no inputs fixed).
    :type param_fixed: list of bools, optional
    :param i: index of the repeated measurement
    :type i: int
    :param repeat_dims: dimension along which the measurements are repeated
    :type repeat_dims: int
    :param repeat_shape: shape of measurements along which to select repeats
    :type repeat_shape: tuple
    :return: list of input quantities, list of uncertainties for single measurement
    :rtype: list[array]. list[array]
    """
    xb = np.zeros(len(x), dtype=object)
    u_xb = np.zeros(len(u_x), dtype=object)
    for j in range(len(x)):
        selected = False
        if param_fixed is not None:
            if param_fixed[j] == True:
                xb[j] = x[j]
                u_xb[j] = u_x[j]
                selected = True
        if not selected:
            index = list(np.ndindex(repeat_shape))[i]
            xb[j] = x[j]
            u_xb[j] = u_x[j]
            for idim in range(len(repeat_dims)):
                repeat_axis = repeat_dims[idim]
                ii = index[idim]
                if len(xb[j].shape) > repeat_axis:
                    if repeat_axis == 0:
                        xb[j] = xb[j][ii]
                        u_xb[j] = u_xb[j][ii]
                    elif repeat_axis == 1:
                        xb[j] = xb[j][:, ii]
                        u_xb[j] = u_xb[j][:, ii]
                    elif repeat_axis == 2:
                        xb[j] = xb[j][:, :, ii]
                        u_xb[j] = u_xb[j][:, :, ii]
                    else:
                        warnings.warn(
                            "The repeat axis is too large to be dealt with by the"
                            "current version of punpy."
                        )
                else:
                    xb[j] = xb[j][ii]
                    u_xb[j] = u_xb[j][ii]

    return xb, u_xb


def nearestPD_cholesky(A, diff=0.001, corr=False, return_cholesky=True):
    """
    Find the nearest positive-definite matrix

    :param A: correlation matrix or covariance matrix
    :type A: array
    :return: nearest positive-definite matrix
    :rtype: array

    Copied and adapted from [1] under BSD license.
    A Python/Numpy port of John D'Errico's `nearestSPD` MATLAB code [2], which
    credits [3].
    [1] https://gist.github.com/fasiha/fdb5cec2054e6f1c6ae35476045a0bbd
    [2] https://www.mathworks.com/matlabcentral/fileexchange/42885-nearestspd
    [3] N.J. Higham, "Computing a nearest symmetric positive semidefinite
    matrix" (1988): https://doi.org/10.1016/0024-3795(88)90223-6
    """

    B = (A + A.T) / 2
    _, s, V = np.linalg.svd(B)

    H = np.dot(V.T, np.dot(np.diag(s), V))

    A2 = (B + H) / 2

    A3 = (A2 + A2.T) / 2

    try:
        return np.linalg.cholesky(A3)
    except:

        spacing = np.spacing(np.linalg.norm(A))

        I = np.eye(A.shape[0])
        k = 1
        while not isPD(A3):
            mineig = np.min(np.real(np.linalg.eigvals(A3)))
            A3 += I * (-mineig * k ** 2 + spacing)
            k += 1

        if corr == True:
            A3 = correlation_from_covariance(A3)
            maxdiff = np.max(np.abs(A - A3))
            if maxdiff > diff:
                raise ValueError(
                    "One of the correlation matrices is not postive definite. "
                    "Correlation matrices need to be at least positive "
                    "semi-definite."
                )
            else:
                warnings.warn(
                    "One of the correlation matrices is not positive "
                    "definite. It has been slightly changed (maximum difference "
                    "of %s) to accomodate our method." % (maxdiff)
                )
                if return_cholesky:
                    return np.linalg.cholesky(A3)
                else:
                    return A3
        else:
            maxdiff = np.max(np.abs(A - A3) / (A3 + diff))
            if maxdiff > diff:
                raise ValueError(
                    "One of the provided covariance matrices is not postive "
                    "definite. Covariance matrices need to be at least positive "
                    "semi-definite. Please check your covariance matrix."
                )
            else:
                warnings.warn(
                    "One of the provided covariance matrix is not positive"
                    "definite. It has been slightly changed (maximum difference of "
                    "%s percent) to accomodate our method." % (maxdiff * 100)
                )
                if return_cholesky:
                    return np.linalg.cholesky(A3)
                else:
                    return A3


def isPD(B):
    """
    Returns true when input is positive-definite, via Cholesky

    :param B: matrix
    :type B: array
    :return: true when input is positive-definite
    :rtype: bool
    """
    try:
        _ = np.linalg.cholesky(B)
        return True
    except np.linalg.LinAlgError:
        return False


def correlation_from_covariance(covariance):
    """
    Convert covariance matrix to correlation matrix

    :param covariance: Covariance matrix
    :type covariance: array
    :return: Correlation matrix
    :rtype: array
    """
    v = np.sqrt(np.diag(covariance))
    outer_v = np.outer(v, v)
    correlation = covariance / outer_v
    correlation[covariance == 0] = 0
    return correlation


def uncertainty_from_covariance(covariance):
    """
    Convert covariance matrix to uncertainty

    :param covariance: Covariance matrix
    :type covariance: array
    :return: uncertainties
    :rtype: array
    """
    return np.sqrt(np.diag(covariance))


def convert_corr_to_cov(corr, u):
    """
    Convert correlation matrix to covariance matrix

    :param corr: correlation matrix
    :type corr: array
    :param u: uncertainties
    :type u: array
    :return: covariance matrix
    :rtype: array
    """
    return u.reshape((-1, 1)) * corr * (u.reshape((1, -1)))


def convert_cov_to_corr(cov, u):
    """
    Convert covariance matrix to correlation matrix

    :param corr: covariance matrix
    :type corr: array
    :param u: uncertainties
    :type u: array
    :return: correlation matrix
    :rtype: array
    """
    return 1 / u.reshape((-1, 1)) * cov / (u.reshape((1, -1)))

  # def fast_positive_definite_inverse(self,m):
    #     cholesky, info = lapack.dpotrf(m)
    #     if info != 0:
    #         raise ValueError('dpotrf failed on input {}'.format(m))
    #     inv, info = lapack.dpotri(cholesky)
    #     if info != 0:
    #         raise ValueError('dpotri failed on input {}'.format(cholesky))
    #     self.upper_triangular_to_symmetric(inv)
    #     return inv


"""
Tests for mc propagation class
"""

import unittest
import numpy as np
import numpy.testing as npt
import punpy.utilities.utilities as util
from punpy.lpu.lpu_propagation import LPUPropagation

"""___Authorship___"""
__author__ = "Pieter De Vis"
__created__ = "14/4/2020"
__maintainer__ = "Pieter De Vis"
__email__ = "pieter.de.vis@npl.co.uk"
__status__ = "Development"


def function(x1, x2):
    return x1 ** 2 - 10 * x2


def Jac_function(x1, x2):
    Jac_x1 = np.diag(2 * x1)
    Jac_x2 = np.diag(-10 * np.ones_like(x2))
    Jac = np.concatenate((Jac_x1, Jac_x2)).T
    return Jac


x1 = np.ones(200) * 10
x2 = np.ones(200) * 30
x1err = np.ones(200)
x2err = 2 * np.ones(200)

xs = np.array([x1, x2])
xerrs = np.array([x1err, x2err])

# below, the higher order Taylor expansion terms have been taken into account, and amount to 2.
yerr_uncorr = 802 ** 0.5 * np.ones(200)
yerr_corr = 2 ** 0.5 * np.ones(200)

# below, the higher order Taylor expansion terms have not been taken into account
yerr_uncorr_1order = 800 ** 0.5 * np.ones(200)
yerr_corr_1order = 0 * np.ones(200)


def functionb(x1, x2):
    return 2 * x1 - x2


x1b = np.ones((20, 3)) * 50
x2b = np.ones((20, 3)) * 30
x1errb = np.ones((20, 3))
x2errb = 2 * np.ones((20, 3))

xsb = np.array([x1b, x2b])
xerrsb = np.array([x1errb, x2errb])

yerr_uncorrb = 8 ** 0.5 * np.ones((20, 3))
yerr_corrb = np.zeros((20, 3))


def functionc(x1, x2, x3):
    return x1 + 4 * x2 - 2 * x3


def Jac_functionc(x1, x2, x3):
    Jac_x1 = np.diag(np.ones_like(x2))
    Jac_x2 = np.diag(4 * np.ones_like(x2))
    Jac_x3 = np.diag(-2 * np.ones_like(x2))
    Jac = np.concatenate((Jac_x1, Jac_x2, Jac_x3)).T
    return Jac


x1c = np.ones(200) * 10
x2c = np.ones(200) * 10
x3c = np.ones(200) * 10

x1errc = 12 * np.ones(200)
x2errc = 5 * np.ones(200)
x3errc = np.zeros(200)

xsc = np.array([x1c, x2c, x3c])
xerrsc = np.array([x1errc, x2errc, x3errc])
corr_c = np.array([[1, 0.9999999, 0], [0.99999999, 1.0, 0], [0.0, 0.0, 1.0]])
yerr_uncorrc = 544 ** 0.5 * np.ones(200)
yerr_corrc = 1024 ** 0.5 * np.ones(200)


def functiond(x1, x2):
    return 2 * x1 - x2, 2 * x1 + x2


x1d = np.ones((5, 3, 2)) * 50
x2d = np.ones((5, 3, 2)) * 30
x1errd = np.ones((5, 3, 2))
x2errd = 2 * np.ones((5, 3, 2))

xsd = [x1d, x2d]
xerrsd = [x1errd, x2errd]
corr_d = np.ones(
    (2, 2)
)  # np.array([[1,0.9999999,0.9999999],[0.99999999,1.,0.99999999],[0.9999999,0.9999999,1.]])

yerr_uncorrd = [8 ** 0.5 * np.ones((5, 3, 2)), 8 ** 0.5 * np.ones((5, 3, 2))]
yerr_corrd = [np.zeros((5, 3, 2)), 16 ** 0.5 * np.ones((5, 3, 2))]


class TestLPUPropagation(unittest.TestCase):
    """
    Class for unit tests
    """

    def test_propagate_random_1D(self):
        prop = LPUPropagation()
        Jx = Jac_function(*xs)
        uf, ucorr = prop.propagate_random(function, xs, xerrs, return_corr=True, Jx=Jx)
        npt.assert_allclose(ucorr, np.eye(len(ucorr)), atol=0.01)
        npt.assert_allclose(uf, yerr_uncorr_1order, rtol=0.01)
        ucov = util.convert_corr_to_cov(ucorr, uf)
        ucorr2 = util.convert_cov_to_corr(ucov, uf)
        npt.assert_allclose(ucorr, ucorr2, atol=0.01)

        uf = prop.propagate_random(
            function, xs, xerrs, corr_between=np.ones((2, 2)), Jx=Jx
        )
        npt.assert_allclose(uf, yerr_corr_1order, atol=0.01)

        uf, ucorr = prop.propagate_random(function, xs, xerrs, return_corr=True, Jx=Jx)
        npt.assert_allclose(uf, yerr_uncorr_1order, rtol=0.01)

        uf = prop.propagate_random(
            function, xs, xerrs, corr_between=np.ones((2, 2)), Jx=Jx
        )
        npt.assert_allclose(uf, yerr_corr_1order, atol=0.01)

    def test_propagate_random_2D(self):
        prop = LPUPropagation(parallel_cores=2)
        ufb, ucorrb, jac_x = prop.propagate_random(
            functionb,
            xsb,
            xerrsb,
            return_corr=True,
            repeat_dims=1,
            return_Jacobian=True,
        )
        npt.assert_allclose(ucorrb, np.eye(len(ucorrb)), atol=0.01)
        npt.assert_allclose(ufb, yerr_uncorrb, rtol=0.01)

        ufb = prop.propagate_random(
            functionb, xsb, xerrsb, corr_between=np.ones((2, 2))
        )
        npt.assert_allclose(ufb, yerr_corrb, atol=0.03)

        ufb = prop.propagate_random(
            functionb,
            xsb,
            xerrsb,
            corr_between=np.ones((2, 2)),
            repeat_dims=1,
            Jx=jac_x,
        )
        npt.assert_allclose(ufb, yerr_corrb, atol=0.03)

    def test_propagate_random_1D_3var(self):
        prop = LPUPropagation()
        Jx = Jac_functionc(*xsc)

        ufc, ucorrc = prop.propagate_random(
            functionc, xsc, xerrsc, return_corr=True, Jx=Jx
        )
        npt.assert_allclose(ucorrc, np.eye(len(ucorrc)), atol=0.01)
        npt.assert_allclose(ufc, yerr_uncorrc, rtol=0.01)

        ufc = prop.propagate_random(functionc, xsc, xerrsc, corr_between=corr_c, Jx=Jx)
        npt.assert_allclose(ufc, yerr_corrc, rtol=0.01)

        ufc = prop.propagate_random(functionc, xsc, xerrsc, corr_between=corr_c, Jx=Jx)
        npt.assert_allclose(ufc, yerr_corrc, rtol=0.01)

    def test_propagate_random_3D_2out(self):
        prop = LPUPropagation()
        ufd, ucorrd, corr_out = prop.propagate_random(
            functiond, xsd, xerrsd, return_corr=True, output_vars=2
        )
        npt.assert_allclose(ucorrd[0], np.eye(len(ucorrd[0])), atol=0.01)
        npt.assert_allclose(ufd, yerr_uncorrd, rtol=0.01)

        ufd, ucorrd, corr_out = prop.propagate_random(
            functiond, xsd, xerrsd, return_corr=True, output_vars=2, repeat_dims=[1, 2]
        )
        npt.assert_allclose(ufd[0], yerr_uncorrd[0], atol=0.01)
        npt.assert_allclose(ufd[1], yerr_uncorrd[1], atol=0.01)
        npt.assert_allclose(corr_out, np.eye(len(corr_out)), atol=0.01)
        npt.assert_allclose(ucorrd[1], np.eye(len(ucorrd[1])), atol=0.01)

        ufd, ucorrd, corr_out = prop.propagate_random(
            functiond, xsd, xerrsd, corr_between=corr_d, return_corr=True, output_vars=2
        )
        npt.assert_allclose(ufd[0], yerr_corrd[0], atol=0.01)
        npt.assert_allclose(ufd[1], yerr_corrd[1], rtol=0.01)

    def test_propagate_systematic_1D(self):
        prop = LPUPropagation()
        Jx = Jac_function(*xs)
        corr = (np.ones_like(x1err) + np.eye(len(x1err))) / 2
        uf, ucorr = prop.propagate_systematic(
            function,
            xs,
            [x1err, None],
            corr_x=[corr, None],
            return_corr=True,
            fixed_corr_var=True,
            Jx=Jx,
        )
        npt.assert_allclose(ucorr, corr, atol=0.0001)

        uf, ucorr = prop.propagate_systematic(
            function, xs, xerrs, return_corr=True, Jx=Jx
        )
        npt.assert_allclose(ucorr, np.ones_like(ucorr), atol=0.01)
        npt.assert_allclose(uf, yerr_uncorr_1order, atol=0.01)

        uf = prop.propagate_systematic(
            function, xs, xerrs, corr_between=np.ones((2, 2)), Jx=Jx
        )
        npt.assert_allclose(uf, yerr_corr_1order, atol=0.01)

    def test_propagate_systematic_2D(self):
        prop = LPUPropagation()
        ufb, ucorrb, jac_x = prop.propagate_systematic(
            functionb, xsb, xerrsb, return_corr=True, return_Jacobian=True
        )
        npt.assert_allclose(ucorrb, np.ones_like(ucorrb), atol=0.01)
        npt.assert_allclose(ufb, yerr_uncorrb, rtol=0.01)

        ufb, ucorrb = prop.propagate_systematic(
            functionb, xsb, xerrsb, return_corr=True, repeat_dims=1
        )
        npt.assert_allclose(ucorrb, np.ones_like(ucorrb), atol=0.01)

        ufb, ucorrb = prop.propagate_systematic(
            functionb, xsb, xerrsb, return_corr=True, corr_axis=0
        )
        npt.assert_allclose(ucorrb, np.ones_like(ucorrb), atol=0.01)

        ufb = prop.propagate_systematic(
            functionb, xsb, xerrsb, corr_between=np.ones((2, 2)), Jx=jac_x
        )
        npt.assert_allclose(ufb, yerr_corrb, atol=0.03)

    def test_propagate_systematic_1D_3var(self):
        prop = LPUPropagation()
        Jx = Jac_functionc(*xsc)

        ufc, ucorrc = prop.propagate_systematic(
            functionc, xsc, xerrsc, return_corr=True, Jx=Jx
        )
        npt.assert_allclose(ucorrc, np.ones_like(ucorrc), atol=0.01)
        npt.assert_allclose(ufc, yerr_uncorrc, rtol=0.01)

        ufc = prop.propagate_systematic(
            functionc, xsc, xerrsc, corr_between=corr_c, Jx=Jx
        )
        npt.assert_allclose(ufc, yerr_corrc, rtol=0.01)

    def test_propagate_systematic_3D_2out(self):
        prop = LPUPropagation()
        ufd, ucorrd, corr_out = prop.propagate_systematic(
            functiond, xsd, xerrsd, return_corr=True, corr_axis=0, output_vars=2
        )
        npt.assert_allclose(ucorrd[0], np.ones_like(ucorrd[0]), atol=0.01)

        ufd, ucorrd, corr_out = prop.propagate_systematic(
            functiond, xsd, xerrsd, return_corr=True, corr_axis=1, output_vars=2
        )
        npt.assert_allclose(ucorrd[1], np.ones_like(ucorrd[1]), atol=0.01)

        ufd, ucorrd, corr_out = prop.propagate_systematic(
            functiond, xsd, xerrsd, return_corr=True, corr_axis=2, output_vars=2
        )
        npt.assert_allclose(ucorrd[0], np.ones_like(ucorrd[0]), atol=0.01)

    def test_propagate_cov_1D(self):
        prop = LPUPropagation()

        cov = [
            util.convert_corr_to_cov(np.eye(len(xerr.flatten())), xerr)
            for xerr in xerrs
        ]

        Jx = Jac_function(*xs)
        uf, ucorr = prop.propagate_cov(function, xs, cov, return_corr=True, Jx=Jx)
        npt.assert_allclose(ucorr, np.eye(len(ucorr)), atol=0.01)
        npt.assert_allclose(uf, yerr_uncorr_1order, rtol=0.01)

        cov = [
            util.convert_corr_to_cov(
                np.ones((len(xerr.flatten()), len(xerr.flatten()))), xerr
            )
            for xerr in xerrs
        ]
        uf, ucorr = prop.propagate_cov(function, xs, cov, return_corr=True, Jx=Jx)
        npt.assert_allclose(uf, yerr_uncorr_1order, atol=0.01)
        npt.assert_allclose(ucorr, np.ones((len(ucorr), len(ucorr))), atol=0.01)

        cov = [util.convert_corr_to_cov(np.eye(len(xerr)), xerr) for xerr in xerrs]
        uf, ucorr = prop.propagate_cov(
            function, xs, cov, return_corr=True, corr_between=np.ones((2, 2)), Jx=Jx
        )
        # ucorr is full of nans because the uncertainties are 0
        npt.assert_allclose(uf, yerr_corr_1order, atol=0.01)

    def test_propagate_cov_2D(self):
        prop = LPUPropagation()

        covb = [
            util.convert_corr_to_cov(np.eye(len(xerrb.flatten())), xerrb)
            for xerrb in xerrsb
        ]
        ufb, ucorrb = prop.propagate_cov(
            functionb, xsb, covb, return_corr=True, Jx_diag=True
        )
        npt.assert_allclose(ucorrb, np.eye(len(ucorrb)), atol=0.01)
        npt.assert_allclose(ufb, yerr_uncorrb, rtol=0.01)

        covb = [
            util.convert_corr_to_cov(
                np.ones((len(xerrb.flatten()), len(xerrb.flatten())))
                + np.eye(len(xerrb.flatten())),
                xerrb,
            )
            for xerrb in xerrsb
        ]
        ufb, ucorrb = prop.propagate_cov(
            functionb, xsb, covb, return_corr=True, Jx_diag=True
        )
        npt.assert_allclose(ufb, yerr_uncorrb * 2 ** 0.5, rtol=0.01)

        covb = [
            util.convert_corr_to_cov(np.eye(len(xerrb.flatten())), xerrb)
            for xerrb in xerrsb
        ]
        ufb, ucorrb = prop.propagate_cov(
            functionb,
            xsb,
            covb,
            return_corr=True,
            corr_between=np.ones((2, 2)),
            Jx_diag=True,
        )
        npt.assert_allclose(ucorrb, np.eye(len(ucorrb)), atol=0.01)
        npt.assert_allclose(ufb, yerr_corrb, atol=0.03)

    def test_propagate_cov_1D_3var(self):
        prop = LPUPropagation()
        Jx = Jac_functionc(*xsc)

        covc = [
            util.convert_corr_to_cov(np.eye(len(xerrc.flatten())), xerrc)
            for xerrc in xerrsc
        ]
        ufc, ucorrc = prop.propagate_cov(functionc, xsc, covc, return_corr=True, Jx=Jx)
        npt.assert_allclose(ucorrc, np.eye(len(ucorrc)), atol=0.01)
        npt.assert_allclose(ufc, yerr_uncorrc, rtol=0.01)

        covc = [
            util.convert_corr_to_cov(
                np.ones((len(xerrc.flatten()), len(xerrc.flatten())))
                + np.eye(len(xerrc)),
                xerrc,
            )
            for xerrc in xerrsc
        ]
        ufc, ucorrc = prop.propagate_cov(functionc, xsc, covc, return_corr=True, Jx=Jx)
        npt.assert_allclose(ufc, yerr_uncorrc * 2 ** 0.5, rtol=0.01)

        covc = [
            util.convert_corr_to_cov(np.eye(len(xerrc.flatten())), xerrc)
            for xerrc in xerrsc
        ]
        ufc, ucorrc = prop.propagate_cov(
            functionc, xsc, covc, return_corr=True, corr_between=corr_c, Jx=Jx
        )
        npt.assert_allclose(ucorrc, np.eye(len(ucorrc)), atol=0.01)
        npt.assert_allclose(ufc, yerr_corrc, rtol=0.01)

    def test_propagate_cov_3D_2out(self):
        prop = LPUPropagation()

        covd = [
            util.convert_corr_to_cov(np.eye(len(xerrd.flatten())), xerrd)
            for xerrd in xerrsd
        ]
        ufd, ucorrd, corr_out = prop.propagate_cov(
            functiond, xsd, covd, return_corr=True, output_vars=2
        )
        npt.assert_allclose(ucorrd[0], np.eye(len(ucorrd[0])), atol=0.01)
        npt.assert_allclose(ufd, yerr_uncorrd, rtol=0.01)

        covd = [
            util.convert_corr_to_cov(
                np.ones((len(xerrd.flatten()), len(xerrd.flatten()))), xerrd
            )
            for xerrd in xerrsd
        ]
        ufd, ucorrd, corr_out = prop.propagate_cov(
            functiond, xsd, covd, return_corr=True, output_vars=2
        )
        npt.assert_allclose(ucorrd[1], np.ones_like(ucorrd[1]), atol=0.01)
        npt.assert_allclose(ufd, yerr_uncorrd, rtol=0.01)

    def test_propagate_syst_corr_2D(self):
        prop = LPUPropagation()

        corrb = [np.eye(len(xerrb[0].flatten())) for xerrb in xerrsb]
        ufb, ucorrb = prop.propagate_systematic(
            functionb,
            xsb,
            xerrsb,
            corr_x=corrb,
            repeat_dims=0,
            corr_axis=1,
            return_corr=True,
        )
        ufb, ucorrb2 = prop.propagate_systematic(
            functionb,
            xsb,
            xerrsb,
            corr_x=["rand", "rand"],
            repeat_dims=0,
            corr_axis=1,
            return_corr=True,
        )
        npt.assert_allclose(ucorrb, np.eye(len(ucorrb)), atol=0.01)
        npt.assert_allclose(ucorrb2, np.eye(len(ucorrb2)), atol=0.01)
        npt.assert_allclose(ufb, yerr_uncorrb, rtol=0.01)

        corrb = [
            np.ones((len(xerrb[0].flatten()), len(xerrb[0].flatten())))
            for xerrb in xerrsb
        ]
        ufb, ucorrb = prop.propagate_systematic(
            functionb,
            xsb,
            xerrsb,
            corr_x=corrb,
            repeat_dims=0,
            corr_axis=1,
            return_corr=True,
        )
        ufb, ucorrb2 = prop.propagate_systematic(
            functionb,
            xsb,
            xerrsb,
            corr_x=["syst", None],
            repeat_dims=0,
            corr_axis=1,
            return_corr=True,
        )
        npt.assert_allclose(ucorrb, np.ones_like(ucorrb), atol=0.01)
        npt.assert_allclose(ucorrb2, np.ones_like(ucorrb2), atol=0.01)
        npt.assert_allclose(ufb, yerr_uncorrb, rtol=0.01)

    def test_propagate_syst_corr_3D_2out(self):
        prop = LPUPropagation()

        corrd = [np.eye(len(xerrd[:, 0, 0].flatten())) for xerrd in xerrsd]
        ufd, ucorrd, corr_out = prop.propagate_systematic(
            functiond,
            xsd,
            xerrsd,
            corr_x=corrd,
            return_corr=True,
            corr_axis=0,
            repeat_dims=[1, 2],
            output_vars=2,
        )
        npt.assert_allclose(ucorrd[0], np.eye(len(ucorrd[0])), atol=0.01)
        npt.assert_allclose(ufd, yerr_uncorrd, rtol=0.01)

        corrd = [
            np.ones((len(xerrd[:, 0, 0].flatten()), len(xerrd[:, 0, 0].flatten())))
            for xerrd in xerrsd
        ]
        ufd, ucorrd, corr_out = prop.propagate_systematic(
            functiond,
            xsd,
            xerrsd,
            corr_x=corrd,
            return_corr=True,
            corr_axis=0,
            repeat_dims=[1, 2],
            output_vars=2,
        )
        npt.assert_allclose(ucorrd[1], np.ones_like(ucorrd[1]), atol=0.01)
        npt.assert_allclose(ufd, yerr_uncorrd, rtol=0.01)


if __name__ == "__main__":
    unittest.main()



"""
Tests for mc propagation class
"""

import unittest
import numpy as np
import numpy.testing as npt
import punpy.utilities.utilities as util
from punpy.mc.mc_propagation import MCPropagation
import time

"""___Authorship___"""
__author__ = "Pieter De Vis"
__created__ = "14/4/2020"
__maintainer__ = "Pieter De Vis"
__email__ = "pieter.de.vis@npl.co.uk"
__status__ = "Development"


def function(x1, x2):
    return x1 ** 2 - 10 * x2 + 30.0


x1 = np.ones(200) * 10
x2 = np.ones(200) * 30
x1err = np.ones(200)
x2err = 2 * np.ones(200)

xs = [x1, x2]
xerrs = [x1err, x2err]

# below, the higher order Taylor expansion terms have been taken into account, and amount to 2.
yerr_uncorr = 802 ** 0.5 * np.ones(200)
yerr_corr = 2 ** 0.5 * np.ones(200)


def functionb(x1, x2):
    return 2 * x1 - x2


x1b = np.ones((20, 30)) * 50
x2b = np.ones((20, 30)) * 30
x1errb = np.ones((20, 30))
x2errb = 2 * np.ones((20, 30))

xsb = np.array([x1b, x2b])
xerrsb = np.array([x1errb, x2errb])

yerr_uncorrb = 8 ** 0.5 * np.ones((20, 30))
yerr_corrb = np.zeros((20, 30))


def functionc(x1, x2, x3):
    return x1 + 4 * x2 - 2 * x3


x1c = np.ones(200) * 10
x2c = np.ones(200) * 10
x3c = np.ones(200) * 10

x1errc = 12 * np.ones(200)
x2errc = 5 * np.ones(200)
x3errc = np.zeros(200)

xsc = np.array([x1c, x2c, x3c])
xerrsc = np.array([x1errc, x2errc, x3errc])
corr_c = np.array([[1, 0.9999999, 0], [0.99999999, 1.0, 0], [0.0, 0.0, 1.0]])
yerr_uncorrc = 544 ** 0.5 * np.ones(200)
yerr_corrc = 1024 ** 0.5 * np.ones(200)


def functiond(x1, x2):
    return 2 * x1 - x2, 2 * x1 + x2


x1d = np.ones((20, 3, 4)) * 50
x2d = np.ones((20, 3, 4)) * 30
x1errd = np.ones((20, 3, 4))
x2errd = 2 * np.ones((20, 3, 4))

xsd = [x1d, x2d]
xerrsd = [x1errd, x2errd]
corr_d = np.ones(
    (2, 2)
)  # np.array([[1,0.9999999,0.9999999],[0.99999999,1.,0.99999999],[0.9999999,0.9999999,1.]])

yerr_uncorrd = [8 ** 0.5 * np.ones((20, 3, 4)), 8 ** 0.5 * np.ones((20, 3, 4))]
yerr_corrd = [np.zeros((20, 3, 4)), 16 ** 0.5 * np.ones((20, 3, 4))]


class TestMCPropagation(unittest.TestCase):
    """
    Class for unit tests
    """

    def test_propagate_random_1D(self):
        prop = MCPropagation(20000, parallel_cores=0)

        uf, ucorr = prop.propagate_random(function, xs, xerrs, return_corr=True)
        npt.assert_allclose(ucorr, np.eye(len(ucorr)), atol=0.06)
        npt.assert_allclose(uf, yerr_uncorr, rtol=0.06)
        ucov = util.convert_corr_to_cov(ucorr, uf)
        ucorr2 = util.convert_cov_to_corr(ucov, uf)
        npt.assert_allclose(ucorr, ucorr2, atol=0.01)

        uf = prop.propagate_random(function, xs, xerrs, corr_between=np.ones((2, 2)))
        npt.assert_allclose(uf, yerr_corr, rtol=0.06)

        uf, ucorr, yvalues, xvalues = prop.propagate_random(
            function, xs, xerrs, return_corr=True, return_samples=True
        )
        npt.assert_allclose(uf, yerr_uncorr, rtol=0.06)

        uf, yvalues, xvalues = prop.propagate_random(
            function, xs, xerrs, corr_between=np.ones((2, 2)), return_samples=True
        )
        npt.assert_allclose(uf, yerr_corr, rtol=0.06)

    def test_propagate_random_2D(self):
        prop = MCPropagation(20000, parallel_cores=0)
        ufb, ucorrb = prop.propagate_random(
            functionb, xsb, xerrsb, return_corr=True, repeat_dims=1
        )
        npt.assert_allclose(ucorrb, np.eye(len(ucorrb)), atol=0.06)
        npt.assert_allclose(ufb, yerr_uncorrb, rtol=0.06)

        ufb = prop.propagate_random(
            functionb, xsb, xerrsb, corr_between=np.ones((2, 2))
        )
        npt.assert_allclose(ufb, yerr_corrb, atol=0.03)

        ufb, yvaluesb, xvaluesb = prop.propagate_random(
            functionb,
            xsb,
            xerrsb,
            corr_between=np.ones((2, 2)),
            return_samples=True,
            repeat_dims=1,
        )
        npt.assert_allclose(ufb, yerr_corrb, atol=0.03)

    def test_propagate_random_1D_3var(self):
        prop = MCPropagation(20000, parallel_cores=1)
        ufc, ucorrc = prop.propagate_random(functionc, xsc, xerrsc, return_corr=True)
        npt.assert_allclose(ucorrc, np.eye(len(ucorrc)), atol=0.06)
        npt.assert_allclose(ufc, yerr_uncorrc, rtol=0.06)

        ufc = prop.propagate_random(functionc, xsc, xerrsc, corr_between=corr_c)
        npt.assert_allclose(ufc, yerr_corrc, rtol=0.06)

        ufc, yvaluesc, xvaluesc = prop.propagate_random(
            functionc, xsc, xerrsc, corr_between=corr_c, return_samples=True
        )
        npt.assert_allclose(ufc, yerr_corrc, rtol=0.06)

    def test_propagate_random_3D_2out(self):
        prop = MCPropagation(20000, parallel_cores=0)
        ufd, ucorrd, corr_out = prop.propagate_random(
            functiond, xsd, xerrsd, return_corr=True, output_vars=2
        )
        npt.assert_allclose(ucorrd[0], np.eye(len(ucorrd[0])), atol=0.06)
        npt.assert_allclose(ufd, yerr_uncorrd, rtol=0.06)

        ufd, ucorrd, corr_out, yvalues, xvalues = prop.propagate_random(
            functiond,
            xsd,
            xerrsd,
            return_corr=True,
            output_vars=2,
            return_samples=True,
            repeat_dims=[2, 1],
        )
        npt.assert_allclose(ucorrd[1], np.eye(len(ucorrd[1])), atol=0.06)

        ufd = prop.propagate_random(
            functiond, xsd, xerrsd, corr_between=corr_d, output_vars=2
        )
        npt.assert_allclose(ufd[0], yerr_corrd[0], atol=0.06)
        npt.assert_allclose(ufd[1], yerr_corrd[1], rtol=0.06)

        ufd, yvaluesd, xvaluesd = prop.propagate_random(
            functiond,
            xsd,
            xerrsd,
            corr_between=corr_d,
            return_samples=True,
            output_vars=2,
        )
        npt.assert_allclose(ufd, yerr_corrd, atol=0.1)

    def test_propagate_systematic_1D(self):
        prop = MCPropagation(30000)

        corr = (np.ones_like(x1err) + np.eye(len(x1err))) / 2
        uf, ucorr = prop.propagate_systematic(
            function,
            xs,
            [x1err, None],
            corr_x=[corr, None],
            return_corr=True,
            fixed_corr_var=True,
        )
        npt.assert_allclose(ucorr, corr, atol=0.0001)

        uf, ucorr = prop.propagate_systematic(function, xs, xerrs, return_corr=True)
        npt.assert_allclose(ucorr, np.ones_like(ucorr), atol=0.06)
        npt.assert_allclose(uf, yerr_uncorr, rtol=0.06)

        uf = prop.propagate_systematic(
            function, xs, xerrs, corr_between=np.ones((2, 2))
        )
        npt.assert_allclose(uf, yerr_corr, rtol=0.06)

    def test_propagate_systematic_2D(self):
        prop = MCPropagation(30000)
        ufb, ucorrb = prop.propagate_systematic(
            functionb, xsb, xerrsb, return_corr=True
        )
        npt.assert_allclose(ucorrb, np.ones_like(ucorrb), atol=0.06)
        npt.assert_allclose(ufb, yerr_uncorrb, rtol=0.06)

        ufb, ucorrb = prop.propagate_systematic(
            functionb, xsb, xerrsb, return_corr=True, repeat_dims=1
        )
        print(ucorrb.shape)
        npt.assert_allclose(ucorrb, np.ones_like(ucorrb), atol=0.06)

        ufb, ucorrb = prop.propagate_systematic(
            functionb, xsb, xerrsb, return_corr=True, corr_axis=0
        )
        print(ucorrb.shape)
        npt.assert_allclose(ucorrb, np.ones_like(ucorrb), atol=0.06)

        ufb = prop.propagate_systematic(
            functionb, xsb, xerrsb, corr_between=np.ones((2, 2))
        )
        npt.assert_allclose(ufb, yerr_corrb, atol=0.03)

    def test_propagate_systematic_1D_3var(self):
        prop = MCPropagation(30000)
        ufc, ucorrc = prop.propagate_systematic(
            functionc, xsc, xerrsc, return_corr=True
        )
        npt.assert_allclose(ucorrc, np.ones_like(ucorrc), atol=0.06)
        npt.assert_allclose(ufc, yerr_uncorrc, rtol=0.06)

        ufc = prop.propagate_systematic(functionc, xsc, xerrsc, corr_between=corr_c)
        npt.assert_allclose(ufc, yerr_corrc, rtol=0.06)

    def test_propagate_systematic_3D_2out(self):
        prop = MCPropagation(30000)
        ufd, ucorrd, corr_out = prop.propagate_systematic(
            functiond, xsd, xerrsd, return_corr=True, corr_axis=0, output_vars=2
        )
        npt.assert_allclose(ucorrd[0], np.ones_like(ucorrd[0]), atol=0.06)

        ufd, ucorrd, corr_out = prop.propagate_systematic(
            functiond, xsd, xerrsd, return_corr=True, corr_axis=1, output_vars=2
        )
        npt.assert_allclose(ucorrd[1], np.ones_like(ucorrd[1]), atol=0.06)

        ufd, ucorrd, corr_out = prop.propagate_systematic(
            functiond, xsd, xerrsd, return_corr=True, corr_axis=2, output_vars=2
        )
        npt.assert_allclose(ucorrd[0], np.ones_like(ucorrd[0]), atol=0.06)

    def test_propagate_cov_1D(self):
        prop = MCPropagation(20000)

        cov = [
            util.convert_corr_to_cov(np.eye(len(xerr.flatten())), xerr)
            for xerr in xerrs
        ]
        uf, ucorr = prop.propagate_cov(function, xs, cov, return_corr=True)
        npt.assert_allclose(ucorr, np.eye(len(ucorr)), atol=0.06)
        npt.assert_allclose(uf, yerr_uncorr, rtol=0.06)

        cov = [
            util.convert_corr_to_cov(
                np.ones((len(xerr.flatten()), len(xerr.flatten()))) + np.eye(len(xerr)),
                xerr,
            )
            for xerr in xerrs
        ]
        uf, ucorr = prop.propagate_cov(function, xs, cov, return_corr=True)
        npt.assert_allclose(uf, yerr_uncorr * 2 ** 0.5, rtol=0.06)

        cov = [
            util.convert_corr_to_cov(np.eye(len(xerr.flatten())), xerr)
            for xerr in xerrs
        ]
        uf, ucorr = prop.propagate_cov(
            function, xs, cov, return_corr=True, corr_between=np.ones((2, 2))
        )
        npt.assert_allclose(ucorr, np.eye(len(ucorr)), atol=0.06)
        npt.assert_allclose(uf, yerr_corr, rtol=0.06)

    def test_propagate_cov_2D(self):
        prop = MCPropagation(20000)

        covb = [
            util.convert_corr_to_cov(np.eye(len(xerrb.flatten())), xerrb)
            for xerrb in xerrsb
        ]
        ufb, ucorrb = prop.propagate_cov(functionb, xsb, covb, return_corr=True)
        npt.assert_allclose(ucorrb, np.eye(len(ucorrb)), atol=0.06)
        npt.assert_allclose(ufb, yerr_uncorrb, rtol=0.06)

        covb = [
            util.convert_corr_to_cov(
                np.ones((len(xerrb.flatten()), len(xerrb.flatten())))
                + np.eye(len(xerrb.flatten())),
                xerrb,
            )
            for xerrb in xerrsb
        ]
        ufb, ucorrb = prop.propagate_cov(functionb, xsb, covb, return_corr=True)
        npt.assert_allclose(ufb, yerr_uncorrb * 2 ** 0.5, rtol=0.06)

        covb = [
            util.convert_corr_to_cov(np.eye(len(xerrb.flatten())), xerrb)
            for xerrb in xerrsb
        ]
        ufb, ucorrb = prop.propagate_cov(
            functionb, xsb, covb, return_corr=True, corr_between=np.ones((2, 2))
        )
        npt.assert_allclose(ucorrb, np.eye(len(ucorrb)), atol=0.06)
        npt.assert_allclose(ufb, yerr_corrb, atol=0.03)

    def test_propagate_cov_1D_3var(self):
        prop = MCPropagation(20000)
        covc = [
            util.convert_corr_to_cov(np.eye(len(xerrc.flatten())), xerrc)
            for xerrc in xerrsc
        ]
        ufc, ucorrc = prop.propagate_cov(functionc, xsc, covc, return_corr=True)
        npt.assert_allclose(ucorrc, np.eye(len(ucorrc)), atol=0.06)
        npt.assert_allclose(ufc, yerr_uncorrc, rtol=0.06)

        covc = [
            util.convert_corr_to_cov(
                np.ones((len(xerrc.flatten()), len(xerrc.flatten())))
                + np.eye(len(xerrc)),
                xerrc,
            )
            for xerrc in xerrsc
        ]
        ufc, ucorrc = prop.propagate_cov(functionc, xsc, covc, return_corr=True)
        npt.assert_allclose(ufc, yerr_uncorrc * 2 ** 0.5, rtol=0.06)

        covc = [
            util.convert_corr_to_cov(np.eye(len(xerrc.flatten())), xerrc)
            for xerrc in xerrsc
        ]
        ufc, ucorrc = prop.propagate_cov(
            functionc, xsc, covc, return_corr=True, corr_between=corr_c
        )
        npt.assert_allclose(ucorrc, np.eye(len(ucorrc)), atol=0.06)
        npt.assert_allclose(ufc, yerr_corrc, rtol=0.06)

    def test_propagate_cov_3D_2out(self):
        prop = MCPropagation(20000)

        covd = [
            util.convert_corr_to_cov(np.eye(len(xerrd.flatten())), xerrd)
            for xerrd in xerrsd
        ]
        ufd, ucorrd, corr_out = prop.propagate_cov(
            functiond, xsd, covd, return_corr=True, corr_axis=0, output_vars=2
        )
        npt.assert_allclose(ucorrd[0], np.eye(len(ucorrd[0])), atol=0.06)
        npt.assert_allclose(ufd, yerr_uncorrd, rtol=0.06)

        covd = [
            util.convert_corr_to_cov(
                np.ones((len(xerrd.flatten()), len(xerrd.flatten()))), xerrd
            )
            for xerrd in xerrsd
        ]
        ufd, ucorrd, corr_out = prop.propagate_cov(
            functiond, xsd, covd, return_corr=True, corr_axis=0, output_vars=2
        )
        npt.assert_allclose(ucorrd[1], np.ones_like(ucorrd[1]), atol=0.06)
        npt.assert_allclose(ufd, yerr_uncorrd, rtol=0.06)

    def test_propagate_syst_corr_2D(self):
        prop = MCPropagation(20000)

        corrb = [np.eye(len(xerrb[0].flatten())) for xerrb in xerrsb]
        ufb, ucorrb = prop.propagate_systematic(
            functionb,
            xsb,
            xerrsb,
            corr_x=corrb,
            repeat_dims=0,
            corr_axis=1,
            return_corr=True,
        )
        ufb, ucorrb2 = prop.propagate_systematic(
            functionb,
            xsb,
            xerrsb,
            corr_x=["rand", "rand"],
            repeat_dims=0,
            corr_axis=1,
            return_corr=True,
        )
        npt.assert_allclose(ucorrb, np.eye(len(ucorrb)), atol=0.06)
        npt.assert_allclose(ucorrb2, np.eye(len(ucorrb2)), atol=0.06)
        npt.assert_allclose(ufb, yerr_uncorrb, rtol=0.06)

        corrb = [
            np.ones((len(xerrb[0].flatten()), len(xerrb[0].flatten())))
            for xerrb in xerrsb
        ]
        ufb, ucorrb = prop.propagate_systematic(
            functionb,
            xsb,
            xerrsb,
            corr_x=corrb,
            repeat_dims=0,
            corr_axis=1,
            return_corr=True,
        )
        ufb, ucorrb2 = prop.propagate_systematic(
            functionb,
            xsb,
            xerrsb,
            corr_x=["syst", None],
            repeat_dims=0,
            corr_axis=1,
            return_corr=True,
        )
        npt.assert_allclose(ucorrb, np.ones_like(ucorrb), atol=0.06)
        npt.assert_allclose(ucorrb2, np.ones_like(ucorrb2), atol=0.06)
        npt.assert_allclose(ufb, yerr_uncorrb, rtol=0.06)

    def test_propagate_syst_corr_3D_2out(self):
        prop = MCPropagation(20000)

        corrd = [np.eye(len(xerrd[:, 0, 0].flatten())) for xerrd in xerrsd]
        ufd, ucorrd, corr_out = prop.propagate_systematic(
            functiond,
            xsd,
            xerrsd,
            corr_x=corrd,
            return_corr=True,
            corr_axis=0,
            repeat_dims=[1, 2],
            output_vars=2,
        )
        npt.assert_allclose(ucorrd[0], np.eye(len(ucorrd[0])), atol=0.06)
        npt.assert_allclose(ufd, yerr_uncorrd, rtol=0.06)

        corrd = [
            np.ones((len(xerrd[:, 0, 0].flatten()), len(xerrd[:, 0, 0].flatten())))
            for xerrd in xerrsd
        ]
        ufd, ucorrd, corr_out = prop.propagate_systematic(
            functiond,
            xsd,
            xerrsd,
            corr_x=corrd,
            return_corr=True,
            corr_axis=0,
            repeat_dims=[1, 2],
            output_vars=2,
        )
        npt.assert_allclose(ucorrd[1], np.ones_like(ucorrd[1]), atol=0.06)
        npt.assert_allclose(ufd, yerr_uncorrd, rtol=0.06)


if __name__ == "__main__":
    unittest.main()



import io
import os
import re

from setuptools import find_packages
from setuptools import setup


def read(filename):
    filename = os.path.join(os.path.dirname(__file__), filename)
    text_type = type(u"")
    with io.open(filename, mode="r", encoding="utf-8") as fd:
        return re.sub(text_type(r":[a-z]+:`~?(.*?)`"), text_type(r"``\1``"), fd.read())


setup(
    #version=versioneer.get_version(),
    #cmdclass=versioneer.get_cmdclass(),
    version='0.17c',
    name="punpy",
    url="https://gitlab.npl.co.uk/eco/eo/punpy",
    license="None",
    author="Pieter De Vis",
    author_email="pieter.de.vis@npl.co.uk",
    description="Propagating UNcertainties in PYthon",
    long_description=read("README.rst"),
    packages=find_packages(exclude=("tests",)),
    install_requires=["numpy", "emcee", "numdifftools", "scipy"],
    extras_require={"dev": ["pre-commit", "tox", "sphinx", "sphinx_rtd_theme"]},
)
